"""teneva_ht_jax.tensors: various useful HT-tensors.

This module contains the collection of functions for explicit construction of
various useful HT-tensors (only random tensor for now).

"""
import jax
import jax.numpy as jnp


def rand(d, n, r, key, a=-1., b=1.):
    """Construct a random HT-tensor from the uniform distribution.

    Args:
        d (int): number of tensor dimensions.
        n (int): mode size of the tensor.
        r (int): TT-ranks of the tensor. It should be a number (if all ranks
            are equal) or list of the length q-1, where q is a number of levels.
        key (jax.random.PRNGKey): jax random key.
        a (float): minimum value for random items of the HT-cores.
        b (float): maximum value for random items of the HT-cores.

    Returns:
        list: HT-tensor.

    """
    q = d.bit_length() # Full number of levels (e.g., d=8 -> q=4)

    if isinstance(r, int):
        r = [r] * (q-1)

    if len(r) != (q-1):
        raise ValueError('Invalid length of ranks list')

    Y = []

    def _rand_level(key, sh):
        key, key_cur = jax.random.split(key)
        Yl = jax.random.uniform(key_cur, sh, minval=a, maxval=b)
        return Yl, key

    # Build the first level (leafs):
    Yl, key = _rand_level(key, sh=(d, n, r[0]))
    Y.append(Yl) # 3D tensor (len, n, r_up)

    # Build the inner levels:
    for k in range(1, q-1):
        dl = 2**(q-k-1) # Length of the current layer
        Yl, key = _rand_level(key, sh=(dl, r[k-1], r[k], r[k-1]))
        Y.append(Yl) # 4D tensor (len, r_down, r_up, r_down)

    # Build the last level (root):
    Yl, key = _rand_level(key, sh=(r[-1], r[-1]))
    Y.append(Yl) # 2D tensor (r_down, r_down)

    return Y



"""teneva_ht_jax.act_one: single HT-tensor operations.

This module contains the basic operations with one HT-tensor (Y), including
"get", etc.

"""
import jax
import jax.numpy as jnp


def get(Y, k):
    """Compute the element of the HT-tensor.

    Args:
        Y (list): d-dimensional HT-tensor.
        k (np.ndarray): the multi-index for the tensor of the length d.

    Returns:
        float: the element of the HT-tensor.

    """
    def body_leaf(q, data):
        i1, i2, G1, G2, G = data
        q = jnp.einsum('r,q,rsq->s', G1[i1], G2[i2], G)
        return None, q

    def body(q, data):
        g1, g2, G = data
        q = jnp.einsum('r,q,rsq->s', g1, g2, G)
        return None, q

    # Compute for the first level (leafs):
    _, q = jax.lax.scan(body_leaf, None,
        (k[0::2], k[1::2], Y[0][0::2], Y[0][1::2], Y[1]))

    # Compute for the inner levels:
    for k in range(1, len(Y)-2):
        _, q = jax.lax.scan(body, None, (q[0::2], q[1::2], Y[k+1]))

    # Compute for the last level (root):
    q = jnp.einsum('r,q,rq->', q[0], q[1], Y[-1])

    return q



"""teneva_ht_jax.vis: visualization methods for tensors.

This module contains the functions for visualization of HT-tensors.

"""
import jax.numpy as jnp


def show(Y):
    """Display mode size and ranks of the given HT-tensor.

    Args:
        Y (list): HT-tensor.

    Todo:
        Add more accurate and informative visualization.

    """
    if not isinstance(Y, list):
        raise ValueError('Invalid HT-tensor')

    d = Y[0].shape[0]

    print(f'HT-tensor  | d={d:-3d}')
    for q, Yl in enumerate(Y):
        print(f'Level: {q+1:-3d} | Shape: {Yl.shape}')



"""teneva_ht_jax.transformation: transformation of HT-tensors.

This module contains the function for transformation of the HT-tensor into full
format.

"""
import jax
import jax.numpy as jnp


def full(Y):
    """Export HT-tensor to the full format.

    Args:
        Y (list): HT-tensor.

    Returns:
        jnp.ndarray: multidimensional array related to the given HT-tensor.

    Note:
         This function can only be used for relatively small tensors, because
         the resulting tensor will have n^d elements and may not fit in memory
         for large dimensions. And this function does not take advantage of
         jax's ability to speed up the code and can be slow, but it should only
         be meaningfully used for tensors of small dimensions.

    """
    # TODO !
    return 42.



"""teneva_ht_jax.gd: build HT-tensor with gradient descent.

This module contains the function "gd_appr", which computes the HT-tensor by
gradient descent method using the provided train dataset.

"""
import jax
import jax.numpy as jnp
import optax
import teneva_ht_jax as tnv
from time import perf_counter as tpc


def gd_appr(Y, I_trn, y_trn, I_vld=None, y_vld=None, epochs=100, batch=100,
            lr=1.E-4, seed=42, log=True, log_trn=True):
    """Build HT-tensor with gradient descent.

    Args:
        Y (list): HT-tensor, which is the initial approximation for algorithm.
        I_trn (np.ndarray): multi-indices for the tensor in the form of array
            of the shape [samples, d], where d is a number of tensor's
            dimensions and samples is a size of the train dataset.
        y_trn (np.ndarray): values of the tensor for multi-indices I_trn in
            the form of array of the shape [samples].
        I_vld (np.ndarray): optional multi-indices for items of validation
            dataset in the form of array of the shape [samples_vld, d], where
            samples_vld is a size of the validation dataset.
        y_vld (np.ndarray): optional values of the tensor for multi-indices
            I_vld of validation dataset in the form of array of the shape
            [samples_vld].
        epochs (int): number of train epochs.
        batch (int): size of the batch for train.
        lr (float): learning rate for training.
        key (jax.random.PRNGKey): jax random key.
        log (bool): if flag is set, then the information about the progress of
            the algorithm will be printed after each epoch.
        log_trn (bool): if flag is set (and log is also set), then the
            accuracy on the train dataset will be presented after each epoch
            (note that for large training datasets this can take a significant
            amount of time, and it may be better to turn off this flag).

    Returns:
        list: HT-tensor, which is the constructed approximation.

    """
    time = tpc()
    rng = jax.random.PRNGKey(seed)
    get = jax.jit(jax.vmap(tnv.get, (None, 0)))
    optim = optax.adam(lr)

    Y = [Yl.copy() for Yl in Y]
    state = optim.init(Y)

    @jax.jit
    def loss(Y, I, y_real):
        y = get(Y, I)
        # jax.debug.print('{v}', v=y)
        l = jnp.mean(jnp.linalg.norm(y_real - y))
        return l

    loss_grad = jax.jit(jax.grad(loss))

    @jax.jit
    def optimize(Y, state, I_cur, y_cur):
        grads = loss_grad(Y, I_cur, y_cur)
        updates, state = optim.update(grads, state)
        Y = jax.tree_util.tree_map(lambda u, y: y + u, updates, Y)
        return Y, state

    for epoch in range(epochs):
        rng, key = jax.random.split(rng)
        perm = jax.random.permutation(key, I_trn.shape[0])
        I_trn_cur = I_trn[perm]
        y_trn_cur = y_trn[perm]

        for j in range(len(I_trn_cur) // batch):
            Y, state = optimize(Y, state,
                I_trn_cur[j * batch:(j+1)*batch],
                y_trn_cur[j * batch:(j+1)*batch])

        if log:
            text = f'# {epoch+1:-3d} | '
            if log_trn:
                y_our = get(Y, I_trn)
                e_trn = jnp.linalg.norm(y_trn - y_our) / jnp.linalg.norm(y_trn)
                text += f'e_trn : {e_trn:-7.1e} | '
            if I_vld is not None and y_vld is not None:
                y_our = get(Y, I_vld)
                e_vld = jnp.linalg.norm(y_vld - y_our) / jnp.linalg.norm(y_vld)
                text += f'e_vld : {e_vld:-7.1e} | '
            text += f't : {tpc()-time:-9.2f}'
            print(text)

    return Y
